//
// Color Image Class
//

#include <cstdio>
#include <iostream>
#include <fstream>
#include <cstring>
#include <cmath>
#include <cassert>

#include "CImage.h"

using namespace std;

static int	quiet_log;

void
CImage_set_quiet()
{
   quiet_log = 1;
}

void
CImage_reset_quiet()
{
   quiet_log = 0;
}

CImage::CImage(int xs, int ys)
{
   assert(xs > 0 || ys > 0);

   xlength = xs;
   ylength = ys;
   xc = xlength / 2;
   yc = ylength / 2;

   r = new int [ys * xs];
   g = new int [ys * xs];
   b = new int [ys * xs];
   if (!r || !g || !b)
   {
      cerr << "Image(): cannot allocate memories" << endl;
      exit(1);
   }
}

CImage::CImage(int xmin, int ymin , int xmax, int ymax)
{
   assert((xmax - xmin) > 0 || (ymax - ymin) > 0);

   xlength = ymax - ymin;
   ylength = xmax - xmin;
   xc = - ymin;
   yc = xmax;

   r = new int [ylength * xlength];
   g = new int [ylength * xlength];
   b = new int [ylength * xlength];
   if (!r || !g || !b)
   {
      cerr << "Image(): cannot allocate memories" << endl;
      exit(1);
   }
}

CImage::~CImage()
{
   delete [] r;
   delete [] g;
   delete [] b;
}

void
CImage::set_size(int xs, int ys)
{
   assert(xs > 0 || ys > 0);

   if (r)  delete [] r;
   if (g)  delete [] g;
   if (b)  delete [] b;
      
   xlength = xs;
   ylength = ys;

   xc = xlength / 2;
   yc = ylength / 2;

   r = new int [ys * xs];
   g = new int [ys * xs];
   b = new int [ys * xs];
   if (!r || !g || !b)
   {
      cerr << "set_size(): cannot allocate memories" << endl;
      exit(1);
   }
}

void
CImage::set_size(int xmin, int ymin, int xmax, int ymax)
{
   assert((xmax - xmin) > 0 || (ymax - ymin) > 0);

   if (r)  delete [] r;
   if (g)  delete [] g;
   if (b)  delete [] b;
      
   xlength = ymax - ymin;
   ylength = xmax - xmin;

   xc = - ymin;
   yc = xmax;

   r = new int [ylength * xlength];
   g = new int [ylength * xlength];
   b = new int [ylength * xlength];
   if (!r || !g || !b)
   {
      cerr << "set_size(): cannot allocate memories" << endl;
      exit(1);
   }
}

void
CImage::set_center(int x, int y)
{
   xc = x;
   yc = y;
}

// rͤθƤӽФ(ΰ賰: 0)
int
CImage::get_pixel_r(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= xlength)
   {
      if (!quiet_log)
	 cerr << "CImage::get_pixel: x is not valid! " << x << endl;
      return 0;
   }
   if (y < 0 || y >= ylength)
   {
      if (!quiet_log)
	 cerr << "CImage::get_pixel: y is not valid! " << y << endl;
      return 0;
   }

   return r[y * xlength + x];

}

// rͤθƤӽФ(ΰ賰: 0)
int
CImage::get_pixel_g(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= xlength)
   {
      if (!quiet_log)
	 cerr << "CImage::get_pixel: x is not valid! " << x << endl;
      return 0;
   }
   if (y < 0 || y >= ylength)
   {
      if (!quiet_log)
	 cerr << "CImage::get_pixel: y is not valid! " << y << endl;
      return 0;
   }

   return g[y * xlength + x];

}

// rͤθƤӽФ(ΰ賰: 0)
int
CImage::get_pixel_b(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= xlength)
   {
      if (!quiet_log)
	 cerr << "CImage::get_pixel: x is not valid! " << x << endl;
      return 0;
   }
   if (y < 0 || y >= ylength)
   {
      if (!quiet_log)
	 cerr << "CImage::get_pixel: y is not valid! " << y << endl;
      return 0;
   }

   return b[y * xlength + x];

}

// ͤθƤӽФ(ΰ賰: 1)
int
CImage::get_pixel2_r(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= xlength)
   {
      if (!quiet_log)
	 cerr << "CImage::get_pixel2: x is not valid! " << x << endl;
      return -1;
   }
   if (y < 0 || y >= ylength)
   {
      if (!quiet_log)
	 cerr << "CImage::get_pixel2: y is not valid! " << y << endl;
      return -1;
   }
   return r[y * xlength + x];
}

// ͤθƤӽФ(ΰ賰: 1)
int
CImage::get_pixel2_g(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= xlength)
   {
      if (!quiet_log)
	 cerr << "CImage::get_pixel2: x is not valid! " << x << endl;
      return -1;
   }
   if (y < 0 || y >= ylength)
   {
      if (!quiet_log)
	 cerr << "CImage::get_pixel2: y is not valid! " << y << endl;
      return -1;
   }
   return g[y * xlength + x];
}

// ͤθƤӽФ(ΰ賰: 1)
int
CImage::get_pixel2_b(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= xlength)
   {
      if (!quiet_log)
	 cerr << "CImage::get_pixel2: x is not valid! " << x << endl;
      return -1;
   }
   if (y < 0 || y >= ylength)
   {
      if (!quiet_log)
	 cerr << "CImage::get_pixel2: y is not valid! " << y << endl;
      return -1;
   }
   return b[y * xlength + x];
}

// ִͤؿ
static double
interpolation(double x0, double y0, double dx, double dy,
	      double d00, double d01, double d10, double d11)
{
   if (d00 >= 0 && d01 >= 0 && d10 >= 0 && d11 >= 0 )
   {
      return (1.0 - dx) * (1.0 - dy) * d00
	   + dx         * (1.0 - dy) * d10
	   + (1.0 - dx) * dy         * d01
	   + dx         * dy         * d11;
   }
   else if (d00 < 0)
   {
      if (d01 < 0 && d10 < 0)
	 return d11;
      else if (d01 < 0 && d11 < 0)
	 return d10;
      else if (d10 < 0 && d11 < 0)
	 return d01;
      else if (d01 < 0)
	 return (1.0 - dx) * d10 + dx * d11;
      else if (d10 < 0)
	 return (1.0 - dy) * d01 + dy * d11;
      else
      {
	 if (!quiet_log)
	    cerr << "get pixel: error ("
		 << x0 << "," << y0 << ") -- ("
		 << d00 << "," << d01 << "," << d10 << "," << d11
		 << ")" << endl;
	 return -1;
      }
   }
   else if (d11 < 0)
   {
      if (d10 < 0 && d01 < 0)
	 return d00;
      else if (d10 < 0)
	 return (1.0 - dx) * d00 + dx * d01;
      else if (d01 < 0)
	 return (1.0 - dy) * d00 + dy * d10;
      else
      {
	 if (!quiet_log)
	    cerr << "get pixel: error ("
		 << x0 << "," << y0 << ") -- ("
		 << d00 << "," << d01 << "," << d10 << "," << d11
		 << ")" << endl;
	 return -1;
      }
   }
   
   if (!quiet_log)
      cerr << "get pixel: error ("
	   << x0 << "," << y0 << ") -- ("
	   << d00 << "," << d01 << "," << d10 << "," << d11
	   << ")" << endl;
   return -1;
}

// Ϣ³ȤƤβͤμФ
double
CImage::get_cpixel_r(double x0, double y0)
{
   double	dx, dy;
   double	lx, ly;
   int		d00, d10, d01, d11;
   int		i, j;
   
   lx = floor(x0);
   ly = floor(y0);
   dx = x0 - lx;
   dy = y0 - ly;

   i = (int)lx;
   j = (int)ly;

   d00 = get_pixel2_r(i  , j  );
   d01 = get_pixel2_r(i  , j+1);
   d10 = get_pixel2_r(i+1, j  );
   d11 = get_pixel2_r(i+1, j+1);

   return interpolation(x0, y0, dx, dy, d00, d01, d10, d11);
}

// Ϣ³ȤƤβͤμФ
double
CImage::get_cpixel_g(double x0, double y0)
{
   double	dx, dy;
   double	lx, ly;
   int		d00, d10, d01, d11;
   int		i, j;
   
   lx = floor(x0);
   ly = floor(y0);
   dx = x0 - lx;
   dy = y0 - ly;

   i = (int)lx;
   j = (int)ly;

   d00 = get_pixel2_g(i  , j  );
   d01 = get_pixel2_g(i  , j+1);
   d10 = get_pixel2_g(i+1, j  );
   d11 = get_pixel2_g(i+1, j+1);

   return interpolation(x0, y0, dx, dy, d00, d01, d10, d11);
}

// Ϣ³ȤƤβͤμФ
double
CImage::get_cpixel_b(double x0, double y0)
{
   double	dx, dy;
   double	lx, ly;
   int		d00, d10, d01, d11;
   int		i, j;
   
   lx = floor(x0);
   ly = floor(y0);
   dx = x0 - lx;
   dy = y0 - ly;

   i = (int)lx;
   j = (int)ly;

   d00 = get_pixel2_b(i  , j  );
   d01 = get_pixel2_b(i  , j+1);
   d10 = get_pixel2_b(i+1, j  );
   d11 = get_pixel2_b(i+1, j+1);

   return interpolation(x0, y0, dx, dy, d00, d01, d10, d11);
}

// Ϣ³ȤƤβͤμФ: (u,v)ɸ
double
CImage::get_im_cpixel_r(double u0, double v0)
{
   double	du, dv;
   double	lu, lv;
   int		d00, d10, d01, d11;
   int		u, v;
   
   lu = floor(u0);
   lv = floor(v0);
   du = u0 - lu;
   dv = v0 - lv;

   u = (int)lu;
   v = (int)lv;

   d00 = get_im_pixel2_r(u  , v  );
   d01 = get_im_pixel2_r(u  , v+1);
   d10 = get_im_pixel2_r(u+1, v  );
   d11 = get_im_pixel2_r(u+1, v+1);

   return interpolation(u0, v0, du, dv, d00, d01, d10, d11);
}

// Ϣ³ȤƤβͤμФ: (u,v)ɸ
double
CImage::get_im_cpixel_g(double u0, double v0)
{
   double	du, dv;
   double	lu, lv;
   int		d00, d10, d01, d11;
   int		u, v;
   
   lu = floor(u0);
   lv = floor(v0);
   du = u0 - lu;
   dv = v0 - lv;

   u = (int)lu;
   v = (int)lv;

   d00 = get_im_pixel2_g(u  , v  );
   d01 = get_im_pixel2_g(u  , v+1);
   d10 = get_im_pixel2_g(u+1, v  );
   d11 = get_im_pixel2_g(u+1, v+1);

   return interpolation(u0, v0, du, dv, d00, d01, d10, d11);
}

// Ϣ³ȤƤβͤμФ: (u,v)ɸ
double
CImage::get_im_cpixel_b(double u0, double v0)
{
   double	du, dv;
   double	lu, lv;
   int		d00, d10, d01, d11;
   int		u, v;
   
   lu = floor(u0);
   lv = floor(v0);
   du = u0 - lu;
   dv = v0 - lv;

   u = (int)lu;
   v = (int)lv;

   d00 = get_im_pixel2_b(u  , v  );
   d01 = get_im_pixel2_b(u  , v+1);
   d10 = get_im_pixel2_b(u+1, v  );
   d11 = get_im_pixel2_b(u+1, v+1);

   return interpolation(u0, v0, du, dv, d00, d01, d10, d11);
}

// ͤν񤭤
void
CImage::put_pixel_r(int x0, int y0, int p)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= xlength) return;
   if (y < 0 || y >= ylength) return;

   r[y * xlength + x] = p;
}

// ͤν񤭤
void
CImage::put_pixel_g(int x0, int y0, int p)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= xlength) return;
   if (y < 0 || y >= ylength) return;

   g[y * xlength + x] = p;
}

// ͤν񤭤
void
CImage::put_pixel_b(int x0, int y0, int p)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= xlength) return;
   if (y < 0 || y >= ylength) return;

   b[y * xlength + x] = p;
}

void
CImage::plot_cross(int x0, int y0, int len, int r0, int g0, int b0)
{
   int  x, y, l;

   l = len / 2;

   for (y = y0 - l; y < y0 + l; y++)
   {
      put_pixel_r(x0, y, r0);
      put_pixel_g(x0, y, g0);
      put_pixel_b(x0, y, b0);
   }

   for (x = x0 - l; x < x0 + l; x++)
   {
      put_pixel_r(x, y0, r0);
      put_pixel_g(x, y0, g0);
      put_pixel_b(x, y0, b0);
   }
}

void
CImage::plot_im_cross(int x0, int y0, int len, int r0, int g0, int b0)
{
   int  x, y, l;

   l = len / 2;

   for (y = y0 - l; y <= y0 + l; y++)
   {
      put_im_pixel2_r(x0, y, r0);
      put_im_pixel2_g(x0, y, g0);
      put_im_pixel2_b(x0, y, b0);
   }

   for (x = x0 - l; x <= x0 + l; x++)
   {
      put_im_pixel2_r(x, y0, r0);
      put_im_pixel2_g(x, y0, g0);
      put_im_pixel2_b(x, y0, b0);
   }
}

void
CImage::clear(int d)
{
   for (int i = 0; i < ylength * xlength; i++)
      r[i] = g[i] = b[i] = d;
}

void
CImage::trim_size()
{
   int  Xmin, Xmax, Ymin, Ymax;
   int  nxl, nyl;
   int  *nr, *ng, *nb;

   Xmin = xlength;
   Ymin = ylength;
   Xmax = 0;
   Ymax = 0;
   for (int y = 0; y < ylength; y++)
      for (int x = 0; x < xlength; x++)
      {
         if (r[y * xlength + x] != 0
	     || g[y * xlength + x] != 0
	     || b[y * xlength + x] != 0)
         {
            if (Xmin > x)  Xmin = x;
            if (Ymin > y)  Ymin = y;
            if (Xmax < x)  Xmax = x;
            if (Ymax < y)  Ymax = y;
         }
      }

   nxl = Xmax - Xmin + 1;
   nyl = Ymax - Ymin + 1;

   if (!quiet_log)
      cerr << "newsize = " << nxl << " x " << nyl << endl;

   nr = new int [nxl * nyl];
   ng = new int [nxl * nyl];
   nb = new int [nxl * nyl];
   int i = 0;
   for (int y = Ymin; y <= Ymax; y++)
      for (int x = Xmin; x <= Xmax; x++)
      {
         nr[i] = r[y * xlength + x];
         ng[i] = g[y * xlength + x];
         nb[i] = b[y * xlength + x];
	 i++;
      }

   delete [] r;
   delete [] g;
   delete [] b;
   r = nr;
   g = ng;
   b = nb;
   xlength = nxl;
   ylength = nyl;
   xc -= Xmin;
   yc -= Ymin;
}

int
save_ppm_file(CImage& img, char *filename)
{
   ofstream     imgfile(filename);
   char         dbuff[3 * img.xlength];
   int		xs, ys, x;

   xs = img.xlength;
   ys = img.ylength;

   if (imgfile.bad()) /* եΥץ */
      return -1;

   /* ޥåʥС */
   imgfile << "P6\n";
   /*  */
   imgfile << "# generated by ppmutil.c\n";
   /* 濴ɸ */
   imgfile << "#@IC " << img.xc << " " << img.yc << "\n";
   /*  */   
   imgfile << xs << " " << ys << "\n";
   /* ٥ */
   imgfile << "255\n";

   x = 0;
   for (int v = 0; v < ys; v++)
   {
      for (int u = 0; u < xs; u++, x++)
      {
         dbuff[3 * u]     = (char)(0x00ff & img.r[x]);
         dbuff[3 * u + 1] = (char)(0x00ff & img.g[x]);
         dbuff[3 * u + 2] = (char)(0x00ff & img.b[x]);
      }
      imgfile.write(dbuff, 3 * xs);
   }

   imgfile.close();

   return 0;
}

int
load_ppm_file(CImage& img, char *filename)
{
   ifstream     imgfile(filename);
   char		buff[BUFSIZ], magic[BUFSIZ];
   char         *dbuff;
   int		height, width;
   int		maxlevel, x;
   int          xc, yc, cflag = 0;      /* 濴ɸ */

   if (imgfile.bad())
      return -1;

   imgfile.getline(buff, BUFSIZ);
   sscanf(buff, "%s", magic);

   imgfile.getline(buff, BUFSIZ);
   while (buff[0] == '#')
   {
      if (!strncmp(buff, "#@IC", 4))
      {
         cflag = 1;
         sscanf(buff + 4, "%d %d", &xc, &yc);
      }
      imgfile.getline(buff, BUFSIZ);
   }

   sscanf(buff, "%d %d", &width, &height);

   if (cflag)
   {
      img.xc = xc;
      img.yc = yc;
   }
   else
   {
      img.xc = width / 2;
      img.yc = height / 2;
   }

   imgfile.getline(buff, BUFSIZ);
   while (buff[0] == '#')
      imgfile.getline(buff, BUFSIZ);
   sscanf(buff, "%d", &maxlevel);

   if (img.xlength != width || img.ylength != height)
      img.set_size(width, height);

   dbuff = new char [width * 3];

   x = 0;
   for (int v = 0; v < height; v++)
   {
      imgfile.read(dbuff, 3 * width);
      for (int u = 0; u < width; u++, x++)
      {
         img.r[x] = (int)(0x00ff & dbuff[3 * u]);
         img.g[x] = (int)(0x00ff & dbuff[3 * u + 1]);
         img.b[x] = (int)(0x00ff & dbuff[3 * u + 2]);
      }
   }

   delete [] dbuff;

   imgfile.close();

   return 0;
}

// 줾β濴·ơim1  im2 Ť͹碌()
int
impose_ppm(CImage& im1, CImage &im2)
{
   int          x, y, x1, y1;
   int          xs, ys;
   int          dr, dg, db;

//   cerr << "im1.xc = " << im1.xc << "\tim2.xc = " << im2.xc << endl;
//   cerr << "im1.yc = " << im1.yc << "\tim2.yc = " << im2.yc << endl;
   xs = im1.xc - im2.xc;
   ys = im1.yc - im2.yc;
   for (y = 0; y < im2.ylength; y++)
   {
      y1 = y + ys;
      if (y1 < 0 || y1 > im1.ylength) continue;
//      cerr << "y = " << y << "\ty1 = " << y1 << endl;

      for (x = 0; x < im2.xlength; x++)
      {
         x1 = x + xs;
         if (x1 < 0 || x1 > im1.xlength) continue;
         dr = im2.get_im_pixel_r(x, y);
         dg = im2.get_im_pixel_g(x, y);
         db = im2.get_im_pixel_b(x, y);
         im1.put_im_pixel_r(x1, y1, dr);
         im1.put_im_pixel_g(x1, y1, dg);
         im1.put_im_pixel_b(x1, y1, db);
      }
   }
   return 0;
}

int
impose_ppm_file(CImage& img, char *filename)
{
   CImage	TMP;
   int		ret;

   ret = load_ppm_file(TMP, filename);
   if (ret < 0)
      return -1;

   impose_ppm(img, TMP);

   return 0;
}

// ƥץ졼

// (2*xs+1, 2*ys+1)礭Υƥץ졼Ȥ

RealCTemplate::RealCTemplate(int xs, int ys)
{
   assert(xs > 0 || ys > 0);

   x_size = xs * 2 + 1;
   y_size = ys * 2 + 1;
   
   xc = xs;
   yc = ys;
   x_min = -xs;
   y_min = -ys;
   x_max = xs;
   y_max = ys;

   r = new double [y_size * x_size];
   g = new double [y_size * x_size];
   b = new double [y_size * x_size];
   if (!r || !g || !b)
   {
      cerr << "RealCTemplate(): cannot allocate memories" << endl;
      exit(1);
   }
}

RealCTemplate::~RealCTemplate()
{
   delete [] r;
   delete [] g;
   delete [] b;
}

void
RealCTemplate::resize(int xs, int ys)
{
   assert(xs > 0 || ys > 0);

   x_size = xs * 2 + 1;
   y_size = ys * 2 + 1;
   
   xc = xs;
   yc = ys;
   x_min = -xs;
   y_min = -ys;
   x_max = xs;
   y_max = ys;

   if (r) delete [] r;
   if (g) delete [] g;
   if (b) delete [] b;
   r = new double [y_size * x_size];
   g = new double [y_size * x_size];
   b = new double [y_size * x_size];
   if (!r || !g || !b)
   {
      cerr << "RealCTemplate.resize(): cannot reallocate memories" << endl;
      exit(1);
   }
}

double
RealCTemplate::get_pixel_r(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size)
   {
      cerr << "Template::get_pixel: x is not valid: " << x << endl;
      return -1;
   }
   if (y < 0 || y >= y_size)
   {
      cerr << "Template::get_pixel: y is not valid: " << y << endl;
      return -1;
   }

//   cerr << "get_pixel: " << data[y * x_size + x] << endl;
   return r[y * x_size + x];
}

double
RealCTemplate::get_pixel_g(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size)
   {
      cerr << "Template::get_pixel: x is not valid: " << x << endl;
      return -1;
   }
   if (y < 0 || y >= y_size)
   {
      cerr << "Template::get_pixel: y is not valid: " << y << endl;
      return -1;
   }

//   cerr << "get_pixel: " << data[y * x_size + x] << endl;
   return g[y * x_size + x];
}

double
RealCTemplate::get_pixel_b(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size)
   {
      cerr << "Template::get_pixel: x is not valid: " << x << endl;
      return -1;
   }
   if (y < 0 || y >= y_size)
   {
      cerr << "Template::get_pixel: y is not valid: " << y << endl;
      return -1;
   }

//   cerr << "get_pixel: " << data[y * x_size + x] << endl;
   return b[y * x_size + x];
}

void
RealCTemplate::put_pixel_r(int x0, int y0, double p)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size) return;
   if (y < 0 || y >= y_size) return;

   r[y * x_size + x] = p;
}

void
RealCTemplate::put_pixel_g(int x0, int y0, double p)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size) return;
   if (y < 0 || y >= y_size) return;

   g[y * x_size + x] = p;
}

void
RealCTemplate::put_pixel_b(int x0, int y0, double p)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size) return;
   if (y < 0 || y >= y_size) return;

   b[y * x_size + x] = p;
}

IntCTemplate::IntCTemplate(int xs, int ys)
{
   assert(xs > 0 || ys > 0);

   x_size = xs * 2 + 1;
   y_size = ys * 2 + 1;
   
   xc = xs;
   yc = ys;
   x_min = -xs;
   y_min = -ys;
   x_max = xs;
   y_max = ys;

   r = new int [y_size * x_size];
   g = new int [y_size * x_size];
   b = new int [y_size * x_size];
   if (!r || !g || !b)
   {
      cerr << "IntCTemplate(): cannot allocate memories" << endl;
      exit(1);
   }
}

IntCTemplate::~IntCTemplate()
{
   delete [] r;
   delete [] g;
   delete [] b;
}

void
IntCTemplate::resize(int xs, int ys)
{
   assert(xs > 0 || ys > 0);

   x_size = xs * 2 + 1;
   y_size = ys * 2 + 1;
   
   xc = xs;
   yc = ys;
   x_min = -xs;
   y_min = -ys;
   x_max = xs;
   y_max = ys;

   if (r) delete [] r;
   if (g) delete [] g;
   if (b) delete [] b;
   r = new int [y_size * x_size];
   g = new int [y_size * x_size];
   b = new int [y_size * x_size];
   if (!r || !g || !b)
   {
      cerr << "IntCTemplate(): cannot allocate memories" << endl;
      exit(1);
   }
}

int
IntCTemplate::get_pixel_r(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size) return -1;
   if (y < 0 || y >= y_size) return -1;

   return r[y * x_size + x];
}

int
IntCTemplate::get_pixel_g(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size) return -1;
   if (y < 0 || y >= y_size) return -1;

   return g[y * x_size + x];
}

int
IntCTemplate::get_pixel_b(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size) return -1;
   if (y < 0 || y >= y_size) return -1;

   return b[y * x_size + x];
}

void
IntCTemplate::put_pixel_r(int x0, int y0, int p)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size) return;
   if (y < 0 || y >= y_size) return;

   r[y * x_size + x] = p;
}

void
IntCTemplate::put_pixel_g(int x0, int y0, int p)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size) return;
   if (y < 0 || y >= y_size) return;

   g[y * x_size + x] = p;
}

void
IntCTemplate::put_pixel_b(int x0, int y0, int p)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size) return;
   if (y < 0 || y >= y_size) return;

   b[y * x_size + x] = p;
}

// ɸ
int
IntCTemplate::get_im_pixel_r(int u0, int v0)
{
   int	u, v;

   u = u0 + xc;
   v = v0 + yc;
   if (u < 0 || u >= x_size) return -1;
   if (v < 0 || v >= y_size) return -1;

   return r[v * x_size + u];
}

int
IntCTemplate::get_im_pixel_g(int u0, int v0)
{
   int	u, v;

   u = u0 + xc;
   v = v0 + yc;
   if (u < 0 || u >= x_size) return -1;
   if (v < 0 || v >= y_size) return -1;

   return g[v * x_size + u];
}

int
IntCTemplate::get_im_pixel_b(int u0, int v0)
{
   int	u, v;

   u = u0 + xc;
   v = v0 + yc;
   if (u < 0 || u >= x_size) return -1;
   if (v < 0 || v >= y_size) return -1;

   return b[v * x_size + u];
}

void
IntCTemplate::put_im_pixel_r(int u0, int v0, int p)
{
   int	u, v;

   u = u0 + xc;
   v = v0 + yc;
   if (u < 0 || u >= x_size) return;
   if (v < 0 || v >= y_size) return;

   r[v * x_size + u] = p;
}

void
IntCTemplate::put_im_pixel_g(int u0, int v0, int p)
{
   int	u, v;

   u = u0 + xc;
   v = v0 + yc;
   if (u < 0 || u >= x_size) return;
   if (v < 0 || v >= y_size) return;

   g[v * x_size + u] = p;
}

void
IntCTemplate::put_im_pixel_b(int u0, int v0, int p)
{
   int	u, v;

   u = u0 + xc;
   v = v0 + yc;
   if (u < 0 || u >= x_size) return;
   if (v < 0 || v >= y_size) return;

   b[v * x_size + u] = p;
}


// ƥץ졼Ȳ
void
set_template(CImage& Im, RealCTemplate& TP, double xc, double yc)
{
   double	ir, ig, ib;
   
   for (int i = TP.xmin(); i <= TP.xmax(); i++)
      for (int j = TP.ymin(); j <= TP.ymax(); j++)
      {
	 ir = Im.get_cpixel_r(xc+i, yc+j);
	 ig = Im.get_cpixel_g(xc+i, yc+j);
	 ib = Im.get_cpixel_b(xc+i, yc+j);
	 if (ir < 0.0)
	 {
	    cerr << "set_template: error! out of bounds" << endl;
	    return;
	 }
	 TP.put_pixel_r(i, j, ir);
	 TP.put_pixel_g(i, j, ig);
	 TP.put_pixel_b(i, j, ib);
      }
}

// ƥץ졼Ȳ
void
set_template(CImage& Im, IntCTemplate& TP, double xc, double yc)
{
   double	ir, ig, ib;
   
   for (int i = TP.xmin(); i <= TP.xmax(); i++)
      for (int j = TP.ymin(); j <= TP.ymax(); j++)
      {
	 ir = Im.get_cpixel_r(xc+i, yc+j);
	 ig = Im.get_cpixel_g(xc+i, yc+j);
	 ib = Im.get_cpixel_b(xc+i, yc+j);
	 if (ir < 0.0)
	 {
	    cerr << "set_template: error! out of bounds" << endl;
	    return;
	 }
	 TP.put_pixel_r(i, j, (int)ir);
	 TP.put_pixel_g(i, j, (int)ig);
	 TP.put_pixel_b(i, j, (int)ib);
      }
}

// ΰ򥳥ԡ
void
copy_area(CImage& src, int u0, int v0,
	  int width, int height, CImage& dst)
{
   assert(dst.width() == width && dst.height() == height);

   for (int v = 0; v < height; v++)
      for (int u = 0; u < width; u++)
      {
	 dst.put_im_pixel_r(u, v, src.get_im_pixel_r(u + u0, v + v0));
	 dst.put_im_pixel_g(u, v, src.get_im_pixel_g(u + u0, v + v0));
	 dst.put_im_pixel_b(u, v, src.get_im_pixel_b(u + u0, v + v0));
      }
}

// 顼Υ
void
CImagetoImage(CImage& src, Image& dst)
{
   int	dr, dg, db;

   assert(src.width() == dst.width() && src.height() == dst.height());

   for (int v = 0; v < src.height(); v++)
      for (int u = 0; u < src.width(); u++)
      {
	 dr = src.get_im_pixel_r(u, v);
	 dg = src.get_im_pixel_g(u, v);
	 db = src.get_im_pixel_b(u, v);
	 dst.put_im_pixel(u, v, (dr + dg + db) / 3);
      }
}
