//
// Image Class
//

#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstring>
#include <cmath>
#include <cassert>

#include "Image.h"

using namespace std;

static int	quiet_log;

void
Image_set_quiet()
{
   quiet_log = 1;
}

void
Image_reset_quiet()
{
   quiet_log = 0;
}

Image::Image(int xs, int ys)
{
   assert(xs > 0 || ys > 0);

   xlength = xs;
   ylength = ys;
   xc = xlength / 2;
   yc = ylength / 2;

   data = new int [ys * xs];
   if (!data)
   {
      cerr << "Image(): cannot allocate memories" << endl;
      exit(1);
   }
}

Image::Image(int xmin, int ymin , int xmax, int ymax)
{
   assert((xmax - xmin) > 0 || (ymax - ymin) > 0);

   xlength = ymax - ymin;
   ylength = xmax - xmin;
   xc = - ymin;
   yc = xmax;

   data = new int [ylength * xlength];
   if (!data)
   {
      cerr << "Image(): cannot allocate memories" << endl;
      exit(1);
   }
}

Image::~Image()
{
   delete [] data;
}

void
Image::set_size(int xs, int ys)
{
   assert(xs > 0 || ys > 0);

   if (data)  delete [] data;
      
   xlength = xs;
   ylength = ys;

   xc = xlength / 2;
   yc = ylength / 2;

   data = new int [ys * xs];
   if (!data)
   {
      cerr << "set_size(): cannot allocate memories" << endl;
      exit(1);
   }
}

void
Image::set_size(int xmin, int ymin, int xmax, int ymax)
{
   assert((xmax - xmin) > 0 || (ymax - ymin) > 0);

   if (data)  delete [] data;
      
   xlength = ymax - ymin;
   ylength = xmax - xmin;

   xc = - ymin;
   yc = xmax;

   data = new int [ylength * xlength];
   if (!data)
   {
      cerr << "set_size(): cannot allocate memories" << endl;
      exit(1);
   }
}

void
Image::set_center(int x, int y)
{
   xc = x;
   yc = y;
}

// ͤθƤӽФ(ΰ賰: 0)
int
Image::get_pixel(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= xlength)
   {
      if (!quiet_log)
	 cerr << "Image::get_pixel: x is not valid! " << x << endl;
      return 0;
   }
   if (y < 0 || y >= ylength)
   {
      if (!quiet_log)
	 cerr << "Image::get_pixel: y is not valid! " << y << endl;
      return 0;
   }

   return data[y * xlength + x];

}

// ͤθƤӽФ(ΰ賰: 1)
int
Image::get_pixel2(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= xlength)
   {
      if (!quiet_log)
	 cerr << "Image::get_pixel2: x is not valid! " << x << endl;
      return -1;
   }
   if (y < 0 || y >= ylength)
   {
      if (!quiet_log)
	 cerr << "Image::get_pixel2: y is not valid! " << y << endl;
      return -1;
   }
   return data[y * xlength + x];
}

// Ϣ³ȤƤβͤμФ
double
Image::get_cpixel(double x0, double y0)
{
   double	dx, dy;
   double	lx, ly;
   int		d00, d10, d01, d11;
   int		i, j;
   
   lx = floor(x0);
   ly = floor(y0);
   dx = x0 - lx;
   dy = y0 - ly;

   i = (int)lx;
   j = (int)ly;

   d00 = get_pixel2(i  , j  );
   d01 = get_pixel2(i  , j+1);
   d10 = get_pixel2(i+1, j  );
   d11 = get_pixel2(i+1, j+1);

   if (d00 >= 0 && d01 >= 0 && d10 >= 0 && d11 >= 0 )
   {
      return (1.0 - dx) * (1.0 - dy) * d00
	   + dx         * (1.0 - dy) * d10
	   + (1.0 - dx) * dy         * d01
	   + dx         * dy         * d11;
   }
   else if (d00 < 0)
   {
      if (d01 < 0 && d10 < 0)
	 return d11;
      else if (d01 < 0 && d11 < 0)
	 return d10;
      else if (d10 < 0 && d11 < 0)
	 return d01;
      else if (d01 < 0)
	 return (1.0 - dx) * d10 + dx * d11;
      else if (d10 < 0)
	 return (1.0 - dy) * d01 + dy * d11;
      else
      {
	 if (!quiet_log)
	    cerr << "get pixel: error ("
		 << x0 << "," << y0 << ") -- ("
		 << d00 << "," << d01 << "," << d10 << "," << d11
		 << ")" << endl;
	 return -1;
      }
   }
   else if (d11 < 0)
   {
      if (d10 < 0 && d01 < 0)
	 return d00;
      else if (d10 < 0)
	 return (1.0 - dx) * d00 + dx * d01;
      else if (d01 < 0)
	 return (1.0 - dy) * d00 + dy * d10;
      else
      {
	 if (!quiet_log)
	    cerr << "get pixel: error ("
		 << x0 << "," << y0 << ") -- ("
		 << d00 << "," << d01 << "," << d10 << "," << d11
		 << ")" << endl;
	 return -1;
      }
   }
   
   if (!quiet_log)
      cerr << "get pixel: error ("
	   << x0 << "," << y0 << ") -- ("
	   << d00 << "," << d01 << "," << d10 << "," << d11
	   << ")" << endl;
   return -1;
}

// ͤν񤭤
void
Image::put_pixel(int x0, int y0, int p)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= xlength) return;
   if (y < 0 || y >= ylength) return;

   data[y * xlength + x] = p;
}

void
Image::plot_cross(int x0, int y0, int len, int data)
{
   int  x, y, l;

   l = len / 2;

   for (y = y0 - l; y < y0 + l; y++)
      put_pixel(x0, y, data);

   for (x = x0 - l; x < x0 + l; x++)
      put_pixel(x, y0, data);
}

void
Image::plot_im_cross(int x0, int y0, int len, int data)
{
   int  x, y, l;

   l = len / 2;

   for (y = y0 - l; y <= y0 + l; y++)
      put_im_pixel2(x0, y, data);

   for (x = x0 - l; x <= x0 + l; x++)
      put_im_pixel2(x, y0, data);
}

void
Image::clear(int d)
{
   for (int i = 0; i < ylength * xlength; i++)
      data[i] = d;
}

void
Image::trim_size()
{
   int  Xmin, Xmax, Ymin, Ymax;
   int  nxl, nyl;
   int  *ndata;

   Xmin = xlength;
   Ymin = ylength;
   Xmax = 0;
   Ymax = 0;
   for (int y = 0; y < ylength; y++)
      for (int x = 0; x < xlength; x++)
      {
         if (data[y * xlength + x] != 0)
         {
            if (Xmin > x)  Xmin = x;
            if (Ymin > y)  Ymin = y;
            if (Xmax < x)  Xmax = x;
            if (Ymax < y)  Ymax = y;
         }
      }

   nxl = Xmax - Xmin + 1;
   nyl = Ymax - Ymin + 1;

   if (!quiet_log)
      cerr << "newsize = " << nxl << " x " << nyl << endl;

   ndata = new int [nxl * nyl];
   int i = 0;
   for (int y = Ymin; y <= Ymax; y++)
      for (int x = Xmin; x <= Xmax; x++)
         ndata[i++] = data[y * xlength + x];

   delete [] data;
   data = ndata;
   xlength = nxl;
   ylength = nyl;
   xc -= Xmin;
   yc -= Ymin;
}

int
save_pgm_file(Image& img, char *filename)
{
   ofstream	imgfile(filename);
   char         dbuff[img.xlength];
   int		xs, ys, x;

   xs = img.xlength;
   ys = img.ylength;

   if (imgfile.bad()) /* եΥץ󥨥顼 */
      return -1;

   /* ޥåʥС */
   imgfile << "P5\n";
   /*  */
   imgfile << "# generated by pgmutil.c\n";
   /* 濴ɸ() */
   imgfile << "#@IC " << img.xc << " " << img.yc << "\n";
   /*  */
   imgfile << xs << " " << ys << "\n";
   /* ٥ */
   imgfile << "255\n";

   x = 0;
   for (int v = 0; v < ys; v++)
   {
      for (int u = 0; u < xs; u++, x++)
         dbuff[u] = (char)(0x00ff & img.data[x]);
      imgfile.write(dbuff, xs);
   }

   imgfile.close();
   return 0;
}

int
load_pgm_file(Image& img, char *filename)
{
   ifstream	imgfile(filename);
   char		buff[BUFSIZ], magic[BUFSIZ];
   char         *dbuff;
   int		height, width;
   int		maxlevel, x;
   int          xc, yc, cflag = 0;      /* 濴ɸ */

   if (imgfile.bad())
      return -1;

   imgfile.getline(buff, BUFSIZ);
   sscanf(buff, "%s", magic);

   imgfile.getline(buff, BUFSIZ);
   while (buff[0] == '#')
   {
      if (!strncmp(buff, "#@IC", 4))
      {
         cflag = 1;
         sscanf(buff + 4, "%d %d", &xc, &yc);
      }
      imgfile.getline(buff, BUFSIZ);
   }

   sscanf(buff, "%d %d", &width, &height);

   if (cflag)
   {
      img.xc = xc;
      img.yc = yc;
   }
   else
   {
      img.xc = width / 2;
      img.yc = height / 2;
   }

   imgfile.getline(buff, BUFSIZ);
   while (buff[0] == '#')
      imgfile.getline(buff, BUFSIZ);
   sscanf(buff, "%d", &maxlevel);

   if (img.xlength != width || img.ylength != height)
      img.set_size(width, height);

   dbuff = new char [width];
   
   x = 0;
   for (int v = 0; v < height; v++)
   {
      imgfile.read(dbuff, width);
      for (int u = 0; u < width; u++, x++) 
         img.data[x] = (int)(0x00ff & dbuff[u]);
   }

   delete [] dbuff;

   imgfile.close();

   return 0;
}

// 줾β濴·ơim1  im2 Ť͹碌()
int
impose_pgm(Image& im1, Image &im2)
{
   int          x, y, x1, y1;
   int          xs, ys;
   int          d;

//   cerr << "im1.xc = " << im1.xc << "\tim2.xc = " << im2.xc << endl;
//   cerr << "im1.yc = " << im1.yc << "\tim2.yc = " << im2.yc << endl;
   xs = im1.xc - im2.xc;
   ys = im1.yc - im2.yc;
   for (y = 0; y < im2.ylength; y++)
   {
      y1 = y + ys;
      if (y1 < 0 || y1 > im1.ylength) continue;
//      cerr << "y = " << y << "\ty1 = " << y1 << endl;

      for (x = 0; x < im2.xlength; x++)
      {
         x1 = x + xs;
         if (x1 < 0 || x1 > im1.xlength) continue;
         d = im2.get_im_pixel(x, y);
         im1.put_im_pixel(x1, y1, d);
      }
   }
   return 0;
}

int
impose_pgm_file(Image& img, char *filename)
{
   Image	TMP;
   int		ret;

   ret = load_pgm_file(TMP, filename);
   if (ret < 0)
      return -1;

   impose_pgm(img, TMP);

   return 0;
}

// ƥץ졼

// (2*xs+1, 2*ys+1)礭Υƥץ졼Ȥ

RealTemplate::RealTemplate(int xs, int ys)
{
   assert(xs > 0 || ys > 0);

   x_size = xs * 2 + 1;
   y_size = ys * 2 + 1;
   
   xc = xs;
   yc = ys;
   x_min = -xs;
   y_min = -ys;
   x_max = xs;
   y_max = ys;

   data = new double [y_size * x_size];
   if (!data)
   {
      cerr << "RealTemplate(): cannot allocate memories" << endl;
      exit(1);
   }
}

RealTemplate::~RealTemplate()
{
   delete [] data;
}

void
RealTemplate::resize(int xs, int ys)
{
   assert(xs > 0 || ys > 0);

   x_size = xs * 2 + 1;
   y_size = ys * 2 + 1;
   
   xc = xs;
   yc = ys;
   x_min = -xs;
   y_min = -ys;
   x_max = xs;
   y_max = ys;

   if (data)
      delete [] data;
   data = new double [y_size * x_size];
   if (!data)
   {
      cerr << "RealTemplate.resize(): cannot reallocate memories" << endl;
      exit(1);
   }
}

double
RealTemplate::get_pixel(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size)
   {
      cerr << "Template::get_pixel: x is not valid: " << x << endl;
      return -1;
   }
   if (y < 0 || y >= y_size)
   {
      cerr << "Template::get_pixel: y is not valid: " << y << endl;
      return -1;
   }

//   cerr << "get_pixel: " << data[y * x_size + x] << endl;
   return data[y * x_size + x];
}

void
RealTemplate::put_pixel(int x0, int y0, double p)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size) return;
   if (y < 0 || y >= y_size) return;

   data[y * x_size + x] = p;
}

double
RealTemplate::get_im_pixel(int u0, int v0)
{
   int	u, v;

   u = u0 + xc;
   v = v0 + yc;
   if (u < 0 || u >= x_size) return -1;
   if (v < 0 || v >= y_size) return -1;

   return data[v * x_size + u];
}

void
RealTemplate::put_im_pixel(int u0, int v0, double p)
{
   int	u, v;

   u = u0 + xc;
   v = v0 + yc;
   if (u < 0 || u >= x_size) return;
   if (v < 0 || v >= y_size) return;

   data[v * x_size + u] = p;
}

// ƥץ졼

IntTemplate::IntTemplate(int xs, int ys)
{
   assert(xs > 0 || ys > 0);

   x_size = xs * 2 + 1;
   y_size = ys * 2 + 1;
   
   xc = xs;
   yc = ys;
   x_min = -xs;
   y_min = -ys;
   x_max = xs;
   y_max = ys;

   data = new int [y_size * x_size];
   if (!data)
   {
      cerr << "IntTemplate(): cannot allocate memories" << endl;
      exit(1);
   }
}

IntTemplate::~IntTemplate()
{
   delete [] data;
}

void
IntTemplate::resize(int xs, int ys)
{
   assert(xs > 0 || ys > 0);

   x_size = xs * 2 + 1;
   y_size = ys * 2 + 1;
   
   xc = xs;
   yc = ys;
   x_min = -xs;
   y_min = -ys;
   x_max = xs;
   y_max = ys;

   if (data)
      delete [] data;
   data = new int [y_size * x_size];
   if (!data)
   {
      cerr << "IntTemplate(): cannot allocate memories" << endl;
      exit(1);
   }
}

int
IntTemplate::get_pixel(int x0, int y0)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size) return -1;
   if (y < 0 || y >= y_size) return -1;

   return data[y * x_size + x];

}

void
IntTemplate::put_pixel(int x0, int y0, int p)
{
   int  x, y;

   x = y0 + xc;
   y = yc - x0;

   if (x < 0 || x >= x_size) return;
   if (y < 0 || y >= y_size) return;

   data[y * x_size + x] = p;
}

int
IntTemplate::get_im_pixel(int u0, int v0)
{
   int	u, v;

   u = u0 + xc;
   v = v0 + yc;
   if (u < 0 || u >= x_size) return -1;
   if (v < 0 || v >= y_size) return -1;

   return data[v * x_size + u];
}

void
IntTemplate::put_im_pixel(int u0, int v0, int p)
{
   int	u, v;

   u = u0 + xc;
   v = v0 + yc;
   if (u < 0 || u >= x_size) return;
   if (v < 0 || v >= y_size) return;

   data[v * x_size + u] = p;
}

// ƥץ졼Ȳ
void
set_template(Image& Im, RealTemplate& TP, double xc, double yc)
{
   double	im;
   
   for (int i = TP.xmin(); i <= TP.xmax(); i++)
      for (int j = TP.ymin(); j <= TP.ymax(); j++)
      {
	 im = Im.get_cpixel(xc+i, yc+j);
	 if (im < 0.0)
	 {
	    cerr << "set_template: error! out of bounds" << endl;
	    return;
	 }
	 TP.put_pixel(i, j, im);
      }
}

// ƥץ졼Ȳ
void
set_template(Image& Im, IntTemplate& TP, double xc, double yc)
{
   double	im;
   
   for (int i = TP.xmin(); i <= TP.xmax(); i++)
      for (int j = TP.ymin(); j <= TP.ymax(); j++)
      {
	 im = Im.get_cpixel(xc+i, yc+j);
	 if (im < 0.0)
	 {
	    cerr << "set_template: error! out of bounds" << endl;
	    return;
	 }
	 TP.put_pixel(i, j, (int)im);
      }
}

// ΰ򥳥ԡ
void
copy_area(Image& src, int u0, int v0,
	  int width, int height, Image& dst)
{
   assert(dst.width() == width && dst.height() == height);

   for (int v = 0; v < height; v++)
      for (int u = 0; u < width; u++)
	 dst.put_im_pixel(u, v, src.get_im_pixel(u + u0, v + v0));
}
