#ifndef _MAT_H_
#define _MAT_H_

#include "nradd.h"
#include "Vec.h"

//
// ޥȥꥯΥ饹
//

class Mat
{
   VecR		*mat;	// ǥΤιԥ٥ȥ
   int		_row;	// Կ
   int		_col;	// 
 public:
   		Mat(int row = DEFAULT_SIZE, int col = DEFAULT_SIZE);
		Mat(double a, double b, double c,
		    double d, double e, double f,
		    double g, double h, double i);
   		Mat(double *d, int row = DEFAULT_SIZE, int col = DEFAULT_SIZE);
   		Mat(const VecC& c1, const VecC& c2, const VecC& c3);
   		Mat(const Mat& M);
   		~Mat();

   void		realloc(int row, int col)
   		{
		   _row = row;
		   _col = col;
		   if (mat != NULL)
		      delete [] mat;
		   
		   mat = new VecR[row];
		   if (mat == NULL)
		   {
		      std::cerr << "### Memory Alloc Error (Mat.realloc) ###\n";
		      abort();
		   }
		   for (int i = 0; i < row; i++)
		   {
		      mat[i].renew(col);
		      for (int j = 0; j < col; j++)
			 mat[i].vec[j] = 0.0;
		   }
		}

   VecR&	operator [] (int i) const // ԥ٥ȥФ
		{
		   assert((i >= 0) && (i < _row));
		   return  mat[i];
		}
   VecC		operator () (int i) const // ٥ȥФ
		{
		   VecC	C(_col);
		   assert((i >= 0) && (i < _col));
		   for (int j = 0; j < _row; j++)
		      C[j] = mat[j].vec[i];
		   return  C;
		}
   Mat&		operator = (const Mat& M) // ޥȥꥯ
		{
		   assert(M._row == _row && M._col == _col);
		   if (&M == this)  return *this;
		   for (int i = 0; i < _row; i++)
		      for (int j = 0; j < _col; j++)
			 mat[i].vec[j] = M[i].vec[j];
		   return *this;
		}
   void         print(std::ostream &s, const char *format);

   Mat&		operator += (const Mat& M);
   Mat&		operator -= (const Mat& M);
   Mat&		operator *= (const Mat& M);
   Mat&		operator *= (const double a);
   Mat&		operator /= (const double a);

   double		norm() const;
   double		sqnorm() const;
   int			row() const { return _row; }
   int			col() const { return _col; }
   void			clear(double data = 0.0);
   void			setcvec(int n, VecC& V);
   friend std::istream&	operator >> (std::istream& s, Mat& M);
   friend std::ostream&	operator << (std::ostream& s, const Mat& M);
   // ޥȥꥯֱ黻
   friend Mat		operator / (const Mat& M, double a);
   friend Mat		operator * (const Mat& M1, const Mat& M2);
   friend Mat		operator * (const Mat& M, double a);
   friend Mat		operator * (double a, const Mat& M);
   friend Mat		operator + (const Mat& M1, const Mat& M2);
   friend Mat		operator - (const Mat& M1, const Mat& M2);
   friend Mat		operator - (const Mat& M);
   friend Mat		operator ~ (const Mat& M);
   friend double	operator , (const Mat& M1, const Mat& M2);
};

// ޥȥꥯ
int	operator == (const Mat& M1, const Mat& M2);
int	operator != (const Mat& M1, const Mat& M2);

// ٥ȥȥޥȥꥯ֤α黻
VecR	operator * (const VecR& RV, const Mat& M);
VecC	operator * (const Mat& M, const VecC& CV);

// ٥ȥ * ԥ٥ȥ => ޥȥꥯ
Mat	operator * (const VecC& CV, const VecR& RV);

// §εչ
Mat	inverse(const Mat& M);

// ̲չ
Mat	gen_inv(const Mat& M);
Mat	trc_gen_inv(const Mat& M, int r);

// 
Mat	ignore_EPS(const Mat& M);
double	det(const Mat& M);
double	det33(const Mat& M);
double	det33EPS(const Mat& M0);

// Υȥ졼
double	trace(const Mat& M);

// ͭͤȸͭ٥ȥ
void	eigen_vector(const Mat& M, Mat& EV, VecC& V);
VecC	min_eigen_vector(const Mat& M);
VecC	min_eigen_vector2(const Mat& M, double *l);

// ðʬ
void    SVD(const Mat& M, Mat& V, VecC& S, Mat& U);

// LUʬˤϢΩ켡β
VecC    SLE(const Mat& A, const VecC& B);

// ñ̹
Mat	id_matrix(int n);

// (exterior product)
Mat	ext_product(const Mat& A, const Mat& B);

// ٥ȥȹȤγ
Mat	operator % (const VecC& v, const Mat& M);

// ٥ȥȤγ
Mat	operator % (const Mat& M, const VecC& V);

// оβ
Mat	sym(const Mat& M);

// ޥȥꥯ3Ĥ٥ȥʬ
void col3(const Mat& M, VecC& u1, VecC& u2, VecC& u3);

// 3x3;ҹ
Mat cofactor33(const Mat& M);

// (ե٥˥Υ1Ȥ)
Mat normalize(const Mat& M);

// гѹ
Mat diag(const Vec& V);
Mat diag33(double a, double b, double c);
Mat diag(const Mat& M);

#endif // _MAT_H_
