#include "nradd.h"

#define MAXIT 100
#define _EPS 3.0e-7
#define FPMIN 1.0e-30

/*
 * sSx[^֐̘AD
 */
double
betacf(double a, double b, double x)
{
   int m,m2;
   double aa,c,d,del,h,qab,qam,qap;

   qab=a+b;
   qap=a+1.0;
   qam=a-1.0;
   c=1.0;
   d=1.0-qab*x/qap;
   if (fabs(d) < FPMIN) d=FPMIN;
   d=1.0/d;
   h=d;
   for (m=1;m<=MAXIT;m++) {
      m2=2*m;
      aa=m*(b-m)*x/((qam+m2)*(a+m2));
      d=1.0+aa*d;
      if (fabs(d) < FPMIN) d=FPMIN;
      c=1.0+aa/c;
      if (fabs(c) < FPMIN) c=FPMIN;
      d=1.0/d;
      h *= d*c;
      aa = -(a+m)*(qab+m)*x/((a+m2)*(qap+m2));
      d=1.0+aa*d;
      if (fabs(d) < FPMIN) d=FPMIN;
      c=1.0+aa/c;
      if (fabs(c) < FPMIN) c=FPMIN;
      d=1.0/d;
      del=d*c;
      h *= del;
      if (fabs(del-1.0) < _EPS) break;
   }
   if (m > MAXIT) nrerror("a or b too big, or MAXIT too small in betacf");
   return h;
}

/*
 * sSx[^֐ I_x(a,b) ԂD
 */
double
betai(double a, double b, double x)
{
   double bt;

   if (x < 0.0 || x > 1.0) nrerror("Bad x in routine betai");
   if (x == 0.0 || x == 1.0) bt=0.0;
   else
      bt=exp(gammln(a+b)-gammln(a)-gammln(b)+a*log(x)+b*log(1.0-x));
   if (x < (a+1.0)/(a+b+2.0))
      return bt*betacf(a,b,x)/a;
   else
      return 1.0-bt*betacf(b,a,1.0-x)/b;
}

/*
 * Fz֐
 */
double
fdist(double f, int n1, int n2)
{
   double	p;

   p = betai(0.5*n2, 0.5*n1, n2/(n2 + n1*f));

   return p;
}

/*
 * Fz̕z֐̋t֐(Fl߂)
 * p0:		Pl
 * n1, n2:	Rx
 */
double
finv(double p0, int n1, int n2)
{
   double	xl, xr, xm;
   double	p;

   xl = 0.0;
   xr = 1e4;

   do
   {
      xm = (xl + xr) / 2.0;
      p = fdist(xm, n1, n2);
      if (p > p0)
	 xl = xm;
      else
	 xr = xm;
   }
   while (fabs(xr - xl)/xm > _EPS);
   
   return xm;
}

#undef	MAXIT 100
#undef	_EPS 3.0e-7
#undef	FPMIN 1.0e-30
