//
// бħ
//
#include <iostream>
#include <fstream>

using namespace std;

#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <sys/types.h>
#include <sys/time.h>

#define MAIN

#define	MAXLEN		256
#define	MAXPOINTS	500

#define	DEFAULT_WIDTH	640
#define	DEFAULT_HEIGHT	480
#define	DEFAULT_RED	1.0
#define	DEFAULT_GREEN	0.0
#define	DEFAULT_BLUE	0.0
#define	DEFAULT_LEN	5
#define	DEFAULT_PLT	1

// ħκɸɤ߹
int
read_points(char *filename, int width, int height, double **x0, double **y0)
{
   ifstream     data_file(filename);
   char		buff[MAXLEN];
   int          N;
   double       x[MAXPOINTS], y[MAXPOINTS];
   int		i;

   if (data_file.bad())
   {
      cerr << "Unable to open " << filename << endl;
      return 0;
   }

   i = 0;
   while (!data_file.eof())
   {
      data_file.getline(buff, MAXLEN);
      sscanf(buff, "(%lf,%lf)", &x[i], &y[i]);
      i++;
      if (i > MAXPOINTS)
	 break;
   }
   data_file.close();

   N = i - 1;
   *x0 = new double [N];
   *y0 = new double [N];

   for (i = 0; i < N; i++)
   {
      (*x0)[i] = x[i];
      (*y0)[i] = y[i];
   }

   return N;
}

void
write_header(int width, int height)
{
   cout << "%!PS-Adobe-2.0 EPSF-2.0" << endl;
   cout << "%%Creator: plot-points" << endl;
   cout << "%%Title: plot-points output" << endl;
   cout << "%%Pages: 1" << endl;
   cout << "%%BoundingBox: -1 -1 " << width + 1 << " " << height + 1 << endl;
   cout << "%%EndComments" << endl;
   cout << "%%Page: 1 1" << endl;
}

void
write_image_circle(int width, int height,
		   double *x0, double *y0, int N,
		   double r, double g, double b, int rad)
{
   double       x, y;
   int		colorflag;

   if (r == g && g == b)
      colorflag = 0;
   else
      colorflag = 1;

   for (int i = 0; i < N; i++)
   {
      if (colorflag)
	 cout << r << " setgray" << endl;
      else
	 cout << r << " " << g << " " << b << " setrgbcolor" << endl;

      x = x0[i];
      y = height - y0[i];

      cout << "newpath" << endl;
      cout << x << " " << y << " " << 2 * rad << " 0 360 arc" << endl;
      cout << "fill" << endl;

      if (colorflag)
	 cout << "0 0 0 setrgbcolor" << endl;
      else
	 cout << "0 setgray" << endl;

      cout << "newpath" << endl;
      cout << x << " " << y << " " << rad << " 0 360 arc" << endl;
      cout << "fill" << endl;
   }
}

void
write_image_cross(int width, int height,
		  double *x0, double *y0, int N,
		  double r, double g, double b, int len)
{
   double       x, y;

   if (r == g && g == b)
      cout << r << " setgray" << endl;
   else
      cout << r << " " << g << " " << b << " setrgbcolor" << endl;

   cout << "[] 0 setdash" << endl;

   cout << "newpath" << endl;
   for (int i = 0; i < N; i++)
   {
      x = x0[i];
      y = height - y0[i];
      cout << x-len << " " << y << " moveto "
	   << x+len << " " << y << " lineto"
	   << endl;
      cout << x << " " << y-len << " moveto "
	   << x << " " << y+len << " lineto"
	   << endl;
   }
   cout << "stroke" << endl;
}

void
write_footer()
{
   cout << "showpage" << endl;
   cout << "%%Trailer" << endl;
}

void
write_ps(int width, int height,
	 double *x0, double *y0, int N,
	 double r, double g, double b, int rol, int cflag)
{
   write_header(width, height);
   if (!cflag)
      write_image_circle(width, height, x0, y0, N, r, g, b, rol);
   else
      write_image_cross(width, height, x0, y0, N, r, g, b, rol);
   write_footer();
}

int
main(int argc, char *argv[])
{
   int          ac;
   double	*x0, *y0;
   double       r, g, b;
   char         Pt_name[80];
   int          width, height, N;
   int		rad_or_len;
   int		cflag;

   if (argc != 2 && argc != 8 && argc != 9)
   {
      cerr << "Usage: plot-points Points_file [width height r g b rad/len [cross_flag]]" << endl;
      cerr << "cross_flag = 0:	plot points using circles" << endl;
      cerr << "             1:	plot points using cross lines" << endl;
      cerr << "DEFAULT:	width = 640, height = 480, color = (1, 0, 0)" << endl;
      cerr << "		cross, length = 5" << endl;
      return -1;
   }

   ac = 0;
   strcpy(Pt_name, argv[++ac]);
   if (argc > 2)
   {
      width = atoi(argv[++ac]);
      height = atoi(argv[++ac]);
      r = atof(argv[++ac]);
      g = atof(argv[++ac]);
      b = atof(argv[++ac]);
      rad_or_len = atoi(argv[++ac]);
      if (argc == 9)
	 cflag = atoi(argv[++ac]);
      else
	 cflag = 0;
   }
   else
   {
      width = DEFAULT_WIDTH;
      height = DEFAULT_HEIGHT;
      r = DEFAULT_RED;
      g = DEFAULT_GREEN;
      b = DEFAULT_BLUE;
      rad_or_len = DEFAULT_LEN;
      cflag = DEFAULT_PLT;
   }

   N = read_points(Pt_name, width, height, &x0, &y0);
   write_ps(width, height, x0, y0, N, r, g, b, rad_or_len, cflag);
}
